#include "aff.h"
#include "aff_et.cpp"
#include "aff_e.cpp"
#include "aff_pi.cpp"
#include "aff_app.cpp"

//-[main]---------------------------------------------------------------------
//
FATHOMi main(int argc, char *argv[])
{
  aff_program art;

  // get system directory (from internal environment variable)
  art.sysdir = get_sysd();

  // process command line option
  if (argc == 2)
    if (art.command_line(argv[1]) == C7EC_OK)
      return C7EC_OK;

  // initial outputting ...
  if (art.chexec(AFFEM_INITIAL_OUTPUT))
  {
    art.f_nf = TRUE;
    art.info_head();
  }

  // error-listing and process-logging
  if (art.chexec(AFFEM_INITIAL_ERROR_LISTING))
    art.enable_errlist("aff_err.txt");
  if (art.chexec(AFFEM_INITIAL_PROCESS_LOGGING))
    art.enable_logfile("aff_log.txt");

  if (art.chexec(AFFEM_NORMAL_PROCESS)) {
  // register  |internal id     |full     |application    |exte-|dire-|
  // types:    |                |name     |type           |nsion|ctory|
  //-----------------+----------------+---------+---------------+-----+-----+
  art.reg_etype(AFF_ELEM_PICTURE,"Picture",AFF_APP_STRETCH,"JPG","./" ,TRUE);
  art.reg_etype(AFF_ELEM_FRAME  ,"Frame"  ,AFF_APP_BORDER ,"GIF","./" ,FALSE);
  art.reg_etype(AFF_ELEM_MATTE  ,"Matte"  ,AFF_APP_AUTO   ,"GIF","./" ,FALSE);
  art.reg_etype(AFF_ELEM_LINER  ,"Liner"  ,AFF_APP_BORDER ,"GIF","./" ,FALSE);
  art.reg_etype(AFF_ELEM_FILLET ,"Fillet" ,AFF_APP_BORDER ,"GIF","./" ,FALSE);
  art.reg_etype(AFF_ELEM_LOGO   ,"Logo"   ,AFF_APP_MASKED ,"JPG","./" ,FALSE);
  art.reg_etype(AFF_ELEM_GRAPHIC,"Graphic",AFF_APP_BLIT   ,"JPG","./" ,FALSE);
  art.reg_etype(AFF_ELEM_TEXT   ,"Text"   ,AFF_APP_TEXT   ,"TXT","./" ,FALSE);

  // get system data (from command-line arguments)
  if (argc > 1)
  {
    FATHOMi i;
    C7STR args;

    for (i = 1; i < argc; i++)
    {
      args.add('&');
      args.add(argv[i]);
    }
    args.add(';');

    art.system_data(args());
  }

  // get data from all available sources
  art.get_data("aff.ini");

  // execute process items
  art.process_items();

  // check security
  art.check_security();

  // build output
  art.build();

  // restrict output
  art.delimit();

  // and do it to it
  art.output(); }

  if (art.chexec(AFFEM_RUN_IO_TEST))
  {
    fstream file;
    char otstdata[50] = "success";
    char itstdata[50] = "error";

    art.tmpll = art.sysdir;
    art.tmpll.add("affio.tst");
    file.open(art.tmpll(), ios::out | ios::trunc);
    file.write(otstdata, strlen(otstdata));

    if (!file.fail())
    {
      file.close();
      art.tmpll = art.sysdir;
      art.tmpll.add("affio.tst");
      file.open(art.tmpll(), ios::in | ios::nocreate);
      file.read(itstdata, strlen(otstdata));
      file.close();
    }


    art.info_head();

    cout << "AFF I/O TEST: " << itstdata << endl
         << "If you see this message then the test probably succeeded!" << endl;

    art.info_foot();
  }

  if (art.chexec(AFFEM_RUN_CGITXT_TEST))
  {
    art.info_head();

    cout << "AFF OUTPUT TEST:" << endl
         << "If you see this message then the test probably succeeded!" << endl;

    art.info_foot();
  }

  if (art.chexec(AFFEM_RUN_CGIJPG_TEST))
  {
    C7GEN cgijpg;
    gen x, y;
    byte *pt = aff_icon_img;

    cgijpg.ini_bmp(77,77,24);

    for (y = 0; y < cgijpg.get_height(); y++)
    {
      for (x = 0; x < cgijpg.get_width(); x++)
      {
        cgijpg.drw_ptrgb(x, y, pt[0], pt[1], pt[2]);
        pt+=3;
      }
    }

    cgi_jpeg(cgijpg, 100);
  }

  return C7EC_OK;
}


//-[aff_program]--------------------------------------------------------------
//
aff_program::aff_program(char *lip)
{
  // initialize main data

  // initialize C7GEN driver
  cw = 0;
  ch = 0;

  // initialize element data
  sz_et  = 0;
  sz_elm = 0;
  sz_pis = 0;

  // initialize delimiter data
  scale = AFF_DEF_SCALE;
  mw    = AFF_DEF_MW;
  mw    = AFF_DEF_MH;
  oq    = AFF_DEF_OQ;
  f_aa  = FALSE; c7gen_affcgi_blur = 0;
  f_sf  = FALSE;

  // if legal ip's have been specified then use them
  if (lip != NULL)
    rip = lip;

  // initialize miscelaneous data
  f_nf = FALSE;
  of.free();

  listing = FALSE;
  logging = FALSE;

  // check execution mode stuff
  chexec();
}

//-[~aff_program]-------------------------------------------------------------
//
aff_program::~aff_program()
{
  if (lf.Length())
  {
    tmpll = sysdir;
    tmpll.add(lf);
    logfile.open(tmpll(), ios::app);

    if (logfile.fail())
    {
      AFF_ERR("LOG FILE", "Unable to open log file (" << lf <<
              ") for append in destructor!");
      lf.free();
    }
    else
    {
      AFF_GET_SYS_TIME;
      logfile << endl << "  ! " <<
              ((status==AFF_ERROR)?
                "Bomb"
                :
                ((status==AFF_WARNING)?"Dirty":"Clean")) <<
              " Termination (" << systime << ')' << endl
              << "----------------------------------------"
              << "----------------------------------------" << endl;
    }
    logfile.close();
  }

  if (f_nf == TRUE)
    info_foot();
}

//-[set some system data]-----------------------------------------------------
//
FATHOMec aff_program::system_data(char *sdat)
{
  sysdat = sdat;
  return C7EC_OK;
}

//-[output]-------------------------------------------------------------------
//
FATHOMec aff_program::output()
{
  AFF_LOG("  -[output]"
          "-----------------------------------"
          "----------------------------------"
          << endl);

  if (status == AFF_ERROR)
  {
    C7STR local_elist;
    C7STR list;
    char the_time[11];
    FATHOMi i;

    if (el.Length()) local_elist.load(el());
    AFF_GET_SYS_TIME;
    memcpy(the_time, systime, 10);
    the_time[10] = '\0';

    if ((i = local_elist.indexOf(the_time)) >= 0)
    {
      list.getpart(local_elist, i, local_elist.Length());
    }

    if (f_nf == TRUE)
    {
      cout << "An error has occured!  CHECK ERROR LISTING FOR DETAILS!" <<  endl;
      if (list.Length())
      {
        cout << "ABBREVIATED ERROR LISTING:"
                "------------------------------------------------------"
             << endl << list << endl
             << "----------------------------------------"
                "----------------------------------------" << endl;
      }
    }

    if ((AFF_CALLTYPE == AFF_CALLTYPE_REMOTE && f_nf == FALSE) ||
        of.Length())
    {
      canvas.ini_bmp(600,400,24);
      canvas.cls_bmp();
      canvas.drw_bot(2,2,canvas.get_rgb(255,255,255),"ERRORS HAVE OCCURED!");
      canvas.drw_bot(2,15,canvas.get_rgb(255,255,255),list());
    }
  }

  AFF_LOG("    @ checking call-type: ");

  if (AFF_CALLTYPE == AFF_CALLTYPE_LOCAL)
  {
    AFF_LOG("local" << endl);
  }
  else
  {
    AFF_LOG("remote (" << CALLER_IP << ")" << endl);

    if (f_nf == FALSE)
    {
      AFF_LOG("    @ saving jpeg through standard output stream" << endl);
      if (cgi_jpeg(canvas, oq) == C7EC_OK)
      {
        AFF_LOG("    @ streaming succesful!" << endl);
      }
      else
      {
        AFF_LOG("    @ streaming failed!" << endl);
        AFF_ERR("OUTPUT", "Could not output binary stream from cgi_jpg()");
      }
    }
  }

  if (of.Length())
  {
    tmpll = sysdir;
    tmpll.add(of);
    AFF_LOG("    @ saving output jpeg file (" << tmpll << ")" << endl);
    if (save_jpeg(canvas,tmpll(),oq) == C7EC_OK)
    {
      AFF_LOG("    @ save successful!" << endl);
    }
    else
    {
      AFF_LOG("    @ save failed!" << endl);
      AFF_ERR("OUTPUT", "Could not save binary output from save_jpg()");
    }
  }

  return C7EC_OK;
}

//-[delimit]------------------------------------------------------------------
//
FATHOMec aff_program::delimit()
{
  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  AFF_LOG("  -[delimit output]"
          "-------------------------------------------------------------"
          << endl);

  if (scale != 1)
  {
    AFF_LOG("    ~ scaling (" << cw << 'x' << ch << ") by " <<
            (scale*100) << "% ");
    cw = (gen)((float)cw * scale);
    ch = (gen)((float)ch * scale);
    AFF_LOG("to (" << cw << 'x' << ch << ")" << endl);
  }

  if (mw && cw > mw)
  {
    AFF_LOG("    ~ lowering (" << cw << 'x' << ch << ") to max width: ");
    ch = (long)(((float)ch) * ( ((float)mw) / ((float)cw)));
    cw = mw;
    AFF_LOG("(" << cw << 'x' << ch << ")" << endl);
  }

  if (mh && ch > mh)
  {
    AFF_LOG("    ~ lowering (" << cw << 'x' << ch << ") to max height: ");
    cw = (long)(((float)cw) * (((float)mh) / ((float)ch)));
    ch = mh;
    AFF_LOG("(" << cw << 'x' << ch << ")" << endl);
  }

  if (cw != canvas.get_width() || ch != canvas.get_height())
  {
    C7GEN fixe;

    AFF_LOG("    ~ stretching canvas (" << canvas.get_width() << 'x' <<
            canvas.get_height() << ") to (" << cw << 'x' << ch << ')' << endl);

    fixe.ini_bmp(canvas.get_width(), canvas.get_height(), 24);
    canvas.blt_norm(fixe,0,0,0,0,canvas.get_width(),canvas.get_height());
    canvas.ini_bmp(cw,ch,24);
    fixe.blt_stretch(canvas,0,0,fixe.get_width(),fixe.get_height(),0,0,cw,ch);
  }

  return C7EC_OK;
}

//-[build]--------------------------------------------------------------------
//
FATHOMec aff_program::build()
{
  FATHOMi i, si;            // index, sub-index
  FATHOMec ecode;           // stores error codes
  float ppi1, ppi2;         // two ppi calculations (width/inch & height/inch)
  C7GEN base;               // image for ppi calculations

  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  AFF_LOG("  -[building output]"
          "------------------------------------------------------------"
          << endl);

  if (sz_elm == 0)
  {
    AFF_ERR("BUILD", "No elements to build!");
    return C7EC_UNABLE_TO_CONTINUE;
  }

  if (et[elm[0].type].calc == FALSE)
  {
    AFF_ERR("BUILD", "Base element type (" << et[elm[0].type].nam <<
            ") cannot be used for calculations!");
    return C7EC_UNABLE_TO_CONTINUE;
  }

  //-[load base image (for ppi calculation)]----------------------------------
  AFF_LOG("    # loading base image (" << elm[0].value() << ")" << endl);
  if (load_image(base,et[elm[0].type].ext,et[elm[0].type].dir,elm[0].value())
      != C7EC_OK)
  {
    return C7EC_UNABLE_TO_CONTINUE;
  }

  //-[calculate pixels per inch]----------------------------------------------
  AFF_LOG("    # calculating ppi: (1)=");
  ppi1 = (float)base.get_width() / (float)elm[0].xval_f[0];
  AFF_LOG(ppi1 << " (2)=");
  ppi2 = (float)base.get_height() / (float)elm[0].xval_f[1];
  AFF_LOG(ppi2);

  if ( (long)(ppi1*AFF_PPI_SENSITIVITY) != (long)(ppi2*AFF_PPI_SENSITIVITY) )
  {
    AFF_WRN("BUILD","Skewed pixels per inch calculation: width/inch=" <<
            ppi1 << "px, height/inch=" << ppi << "px");
    ppi = (ppi1+ppi2) / 2;
    AFF_LOG("!? Averaging=" << ppi << endl);
  }
  else
  {
    if (ppi2 > ppi1)
      ppi = ppi2;
    else
      ppi = ppi1;

    AFF_LOG(" Setting=" << ppi << endl);
  }
  base.kll();

  //-[squeeze out the canvas]-------------------------------------------------
  cw = 0;
  ch = 0;
  i = 0;
  while (i < sz_elm && status != AFF_ERROR)
  {
    switch(et[elm[i].type].app)
    {
      case AFF_APP_PAINT:
      case AFF_APP_TILE:
      case AFF_APP_AUTO:
      case AFF_APP_BORDER:
      case AFF_APP_STRETCH:
      {
        // convert inches to pixels
        si = 0;
        while (si < AFF_MAX_XVARS)
        {
          elm[i].xval_i[si] = (gen)(elm[i].xval_f[si] * ppi);
          if (elm[i].xval_i[si] < 1) elm[i].xval_i[si] = 1;
          si++;
        }
        if (i)
        {
          cw += (elm[i].ls() + elm[i].rs());
          ch += (elm[i].ts() + elm[i].bs());
        }
        else
        {
          cw += elm[i].w();
          ch += elm[i].h();
        }
      } break;
      case AFF_APP_BLIT:
      case AFF_APP_MASKED:
      case AFF_APP_TEXT:
      {
      } break;
      default:
        break;
    }
    i++;  // have to do this or we'll never get home to kansas
  }

  AFF_LOG("    # stretching canvas " << cw << 'x' << ch << endl);
  canvas.ini_bmp(cw,ch,24);

  //-[begin the drawing]------------------------------------------------------

  AFF_LOG("    # begin drawing base elements " << endl);

  x1 = y1 = 0;
  x2 = cw - 1;
  y2 = ch - 1;
  ecode = C7EC_OK;

  i = sz_elm - 1;
  while (i >= 0 && ecode == C7EC_OK)
  {
    switch(et[elm[i].type].app)
    {
      case AFF_APP_PAINT:     ecode = apply_paint(elm[i]);      break;
      case AFF_APP_TILE:      ecode = apply_tile(elm[i]);       break;
      case AFF_APP_AUTO:      ecode = apply_auto(elm[i]);       break;
      case AFF_APP_BORDER:    ecode = apply_border(elm[i]);     break;
      case AFF_APP_STRETCH:   ecode = apply_stretch(elm[i]);    break;
      default:
        break;
    }
    i--;
  }

  AFF_LOG("    # begin drawing special elements " << endl);

  i = 0;
  while (i < sz_elm && ecode == C7EC_OK)
  {
    switch(et[elm[i].type].app)
    {
      case AFF_APP_BLIT:      ecode = apply_blit(elm[i]);       break;
      case AFF_APP_MASKED:    ecode = apply_masked(elm[i]);     break;
      case AFF_APP_TEXT:      ecode = apply_text(elm[i]);       break;
      default:
        break;
    }
    i++;
  }

  if (ecode != C7EC_OK)
  {
    AFF_LOG("    # canvas drawing aborted (" << ecode << ")" << endl);
    return C7EC_UNABLE_TO_CONTINUE;
  }

  AFF_LOG("    # finished building canvas!" << endl);

  return C7EC_OK;
}

//-[process_items]------------------------------------------------------------
//
FATHOMec aff_program::load_image(C7GEN &dst, C7STR &ext, C7STR &dir, char *filename)
{
  C7STR fn;
  C7STR ofn;
  FATHOMec ecode;
  ofn.is(filename);

  if (dir.Length() && ofn.indexOf('/') < 0 && ofn.indexOf('\\') < 0)
  {
    fn = dir;
    fn.add(ofn);
  }
  else
    fn = ofn;

  if (fn.indexOf('.') < 0)
  {
    fn.add('.');
    fn.add(ext);
  }

  AFF_LOG("      $ loading image [" << fn << "] ... ");

  dst.kll();

  fn.case_flag(0);

  if (fn.indexOf(".JPG") >= 0) {
    ecode = load_jpeg(dst, fn());
  } else if (fn.indexOf(".PCX") >= 0) {
    ecode = load_pcx(dst, fn());
  } else if (fn.indexOf(".GIF") >= 0) {
    ecode = load_gif(dst, fn());
  } else if (fn.indexOf(".BMP") >= 0) {
    ecode = load_bmp(dst, fn());
  } else
  {
    AFF_LOG("FAIL: unrecognized extension" << endl);
    AFF_ERR("LOAD IMG", "Bad file extension \"" << fn << "\"!");
    return C7EC_UNABLE_TO_CONTINUE;
  }

  if (ecode != C7EC_OK || dst.get_width() <= 0 || dst.get_height() <= 0)
  {
    AFF_LOG("FAIL: load error" << endl);
    AFF_ERR("LOAD IMG", "Unable to load image \"" << fn << "\" (" << ecode << ")!");
    return C7EC_UNABLE_TO_CONTINUE;
  }

  AFF_LOG("OK (" << dst.get_width() << 'x' << dst.get_height() << ')' << endl);

  return C7EC_OK;
}

//-[process_items]------------------------------------------------------------
//
FATHOMec aff_program::process_items()
{
  FATHOMi i, si;
  FATHOMi het;
  uint32 id;

  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  AFF_LOG("  -[processing items]"
          "-----------------------------------------------------------"
          << endl);

  i = 0;
  while (i < sz_pis && status != AFF_ERROR)
  {
    switch (pis[i].id)
    {
      case AFF_PI_ADD: //-----------------------------------------[add element
      {
        AFF_LOG("    > adding element " << (sz_elm+1) << ": ");
        if (sz_elm >= AFF_MAX_ELEMENTS)
        {
          AFF_LOG("FAIL: maxed out" << endl);
          AFF_ERR("EXEC PI","Maximum number of elements reached (" <<
                  AFF_MAX_ELEMENTS << ")!");
          break;
        }
        else if (pis[i].value[0].Length() != 3)
        {
          AFF_LOG("FAIL: bad type name" << endl);
          AFF_ERR("EXEC PI","Bad element type name (" <<
                  pis[i].value[0] << ")!");
          break;
        }

        id = AFF_EID(pis[i].value[0][0],pis[i].value[0][1],pis[i].value[0][2]);        
        if ((het = handle_etype(id)) >= sz_et)
        {
          AFF_LOG("FAIL: unknown type" << endl);
          AFF_ERR("EXEC PI","Unknown element type for addition (" <<
                  pis[i].value[0] << ")!");
          break;
        }
        elm[sz_elm].type = het;
        AFF_LOG("ID[" << id << ']');

        elm[sz_elm].value = pis[i].value[1];
        AFF_LOG("V[" << pis[i].value[1] << ']');

        switch(et[het].app)
        { //_[beg switch]_____________________________________________________
          case AFF_APP_PAINT:
          case AFF_APP_TILE:
          case AFF_APP_AUTO:
          case AFF_APP_BORDER:
          case AFF_APP_STRETCH:
          {
            for (si = 2; pis[i].value[si].Length() && si-2 < AFF_MAX_XVARS; si++)
            {
              elm[sz_elm].xval_f[si-2] = atof(pis[i].value[si].data());
              AFF_LOG('X' << (si-1) << '[' << elm[sz_elm].xval_f[si-2] << ']');
              elm[sz_elm].numx++;
            }
            for (si-=2; si < AFF_MAX_XVARS; si++)
            {
              elm[sz_elm].xval_f[si] = elm[sz_elm].xval_f[0];
              elm[sz_elm].numx++;
            }
          } break;
          case AFF_APP_BLIT:
          case AFF_APP_MASKED:
          case AFF_APP_TEXT:
          {
            for (si = 2; pis[i].value[si].Length() && si-2 < AFF_MAX_XVARS; si++)
            {
              elm[sz_elm].xval_i[si-2] = strtol(pis[i].value[si].data(),NULL,0);
              AFF_LOG('X' << (si-1) << '[' << elm[sz_elm].xval_i[si-2] << ']');
              elm[sz_elm].numx++;
            }
            for (si-=2; si < AFF_MAX_XVARS; si++)
            {
              elm[sz_elm].xval_i[si] = elm[sz_elm].xval_i[0];
              elm[sz_elm].numx++;
            }
          } break;
        } //_[end switch]_____________________________________________________
        
        AFF_LOG(endl);

        sz_elm++;
      } break;

      case AFF_PI_DIR: //-----------------------[change element type directory
      {
        AFF_LOG("    > element type directory: ");
        id = AFF_EID(pis[i].value[0][0],pis[i].value[0][1],pis[i].value[0][2]);
        if (pis[i].value[0].Length() != 3 || (het = handle_etype(id)) >= sz_et)
        {
          AFF_LOG("FAIL: unknown type" << endl);
          AFF_ERR("EXEC PI","Unknown element type for directory change (" <<
                  pis[i].value[0] << ")!");         
          break;
        }
        et[het].dir = pis[i].value[1];
        AFF_LOG(pis[i].value[0] << "_dir = " << pis[i].value[1] << endl);
      } break;

      case AFF_PI_EXT: //-----------------------[change element type extension
      {
        AFF_LOG("    > element type extension: ");
        id = AFF_EID(pis[i].value[0][0],pis[i].value[0][1],pis[i].value[0][2]);
        if (pis[i].value[0].Length() != 3 || (het = handle_etype(id)) >= sz_et)
        {
          AFF_ERR("EXEC PI","Unknown element type for extension change (" <<
                  pis[i].value[0] << ")!");         
          break;
        }
        et[het].ext = pis[i].value[1];
        AFF_LOG(pis[i].value[0] << "_ext = " << pis[i].value[1] << endl);
      } break;

      case AFF_PI_SCAL: //----------------------------------------------[scale
      { scale = (float)strtol(pis[i].value[0](),NULL,10);
        scale /= 100;
        AFF_LOG("    > scale = ");
        if (scale < AFF_MIN_SCALE || scale > AFF_MAX_SCALE)
        {
          AFF_LOG("<invalid> / restoring default (" << AFF_DEF_SCALE << ')');
          AFF_WRN("EXEC PI", "Scale is invalid (" << scale <<
                  "), keeping default (" << AFF_DEF_SCALE << ')');
          scale = AFF_DEF_SCALE;
        }
        AFF_LOG((scale*100) << '%' << endl);
      } break;

      case AFF_PI_MW: //----------------------------------------[maximum width
      { mw = strtol(pis[i].value[0](),NULL,10);
        AFF_LOG("    > max width = ");
        if (mw < AFF_MIN_MW || mw > AFF_MAX_MW)
        {
          AFF_LOG("<invalid> / restoring default (" << AFF_DEF_MW << ')' <<
                  endl);
          AFF_WRN("EXEC PI", "Maximum width is invalid (" << mw <<
                  "), keeping default (" << AFF_DEF_MW << ')');
          mw = AFF_DEF_MW;
          break;
        }
        AFF_LOG(mw << endl);
      } break;

      case AFF_PI_MH: //---------------------------------------[maximum height
      { mh = strtol(pis[i].value[0](),NULL,10);
        AFF_LOG("    > max height = ");
        if (mh < AFF_MIN_MH || mh > AFF_MAX_MH)
        {
          AFF_LOG("<invalid> / restoring default (" << AFF_DEF_MH << ')' <<
                  endl);
          AFF_WRN("EXEC PI", "Maximum width is invalid (" << mh <<
                  "), keeping default (" << AFF_DEF_MH << ')');
          mh = AFF_DEF_MH;
          break;
        }
        AFF_LOG(mh << endl);
      } break;

      case AFF_PI_OQ: //---------------------------------------[output quality
      { oq = strtol(pis[i].value[0](),NULL,10);
        AFF_LOG("    > output quality = ");
        if (oq > AFF_MAX_OQ || oq < AFF_MIN_OQ)
        {
          AFF_LOG("<invalid> / restoring default (" << AFF_DEF_OQ << ')' <<
                  endl);
          AFF_WRN("EXEC PI", "Output quality is invalid (" << oq <<
                  "), keeping default (" << AFF_DEF_OQ << ')');
          oq = AFF_DEF_OQ;
          break;
        }

        AFF_LOG(oq << endl);
      } break;

      case AFF_PI_F_AA: //-----------------------------------[flag: anti-alias
      {
        AFF_LOG("    > anti-alias = ");
        if (pis[i].value[0][0] == '1' ||
            pis[i].value[0][0] == 't' ||
            pis[i].value[0][0] == 'y' ||
            pis[i].value[0] == "on")
        {
          c7gen_affcgi_blur = 1;
          AFF_LOG("on" << endl);
          f_aa = TRUE;
        }
        else
        {
          c7gen_affcgi_blur = 0;
          AFF_LOG("off" << endl);
          f_aa = FALSE;
        }
      } break;

      case AFF_PI_F_SF: //----------------------------------[flag: scale first
      {
        AFF_LOG("    > scale first = ");
        if (pis[i].value[0][0] == '1' ||
            pis[i].value[0][0] == 't' ||
            pis[i].value[0][0] == 'y' ||
            pis[i].value[0] == "on")
        {
          AFF_LOG("on" << endl);
          f_sf = TRUE;
        }
        else
        {
          AFF_LOG("off" << endl);
          f_sf = FALSE;
        }
      } break;

      case AFF_PI_RIP:  // ---------------------------------------[required ip
      {
        if (rip.Length())
          rip.add(pis[i].value[0]);
        else
          rip = pis[i].value[0];
        AFF_LOG("    > required ip = " << rip << endl);
      } break;

      case AFF_PI_INC:  // --------------------------------------[include file
      { // we already processed the 'inc's inside "get_data"
      } break;

      case AFF_PI_F_NF: // ---------------------------------[flag: output info
      {
        AFF_LOG("    > output info = ");

        if (pis[i].value[0][0] == '1' ||
            pis[i].value[0][0] == 't' ||
            pis[i].value[0][0] == 'y' ||
            pis[i].value[0] == "on")
        {
          info_head();
          AFF_LOG("on" << endl);
          f_nf = TRUE;
        }
        else if (!chexec(AFFEM_INITIAL_OUTPUT))
        {
          AFF_LOG("off" << endl);
          f_nf = FALSE;
        }
      } break;

      case AFF_PI_LF: //---------------------------------------------[log file
      { enable_logfile(pis[i].value[0]());
      } break;

      case AFF_PI_OF: //------------------------------------------[output file
      { of = pis[i].value[0];
        AFF_LOG("    > output file = \"" << of << '"' << endl);
      } break;

      case AFF_PI_EL: //----------------------------------------[error listing
      { AFF_LOG("    > error listing = \"" << pis[i].value[0]() << '"' << endl);
        enable_errlist(pis[i].value[0]());
      } break;

      case AFF_PI_SYSD: //-----------------------------------[system directory
      { sysdir = pis[i].value[0];
        AFF_LOG("    > system directory = \"" << sysdir << "\"" << endl);
      } break;
      default:
      {
        AFF_LOG("    > unknown process item identification: " <<
                pis[i].id << '(' <<
                ((char)AFF_PIB1(pis[i].id)) << ((char)AFF_PIB3(pis[i].id)) <<
                ((char)AFF_PIB3(pis[i].id)) << ((char)AFF_PIB4(pis[i].id)) <<
                ')');
        AFF_WRN("EXEC PI","Unknown process item identification: " <<
                pis[i].id << '(' <<
                ((char)AFF_PIB1(pis[i].id)) << ((char)AFF_PIB3(pis[i].id)) <<
                ((char)AFF_PIB3(pis[i].id)) << ((char)AFF_PIB4(pis[i].id)) <<
                ')');
      } break;
    }
    i++;
  }

  return C7EC_OK;
}

//-[get_data]-----------------------------------------------------------------
//
FATHOMec aff_program::get_data(char *inifile)
{
  C7STR data;           // place to store incoming data
  long clen;            // for storing CONTENT_LENGTH
  FATHOMec ecode;       // for temporary FATHOM error codes
  FATHOMi si,i;         // index variable
  bool anyinp;          // used to check if we find ANY input files
  C7STR fullfn;         // full path and filename of inifn
  C7STR tmp;            //

  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  AFF_LOG("  -[getting data]"
          "---------------------------------------------------------------"
          << endl);

  //-[initialization file]----------------------------------------------------
    tmp.is(inifile);
    if (sysdir.Length() && tmp.indexOf('/') < 0  && tmp.indexOf('\\') < 0)
      fullfn = sysdir;
    else if (tmp.indexOf('/') < 0 && tmp.indexOf('\\') < 0)
      fullfn = "\\";
    else
      fullfn = "";
    fullfn.add(inifile);

    AFF_LOG("    * reading ini file (" << fullfn << ") ... ");

    if ((ecode = data.load(fullfn)) == C7EC_OK)
    {
      AFF_LOG("OKAY!" << endl);
    }
    else
    {
      AFF_LOG("FAIL!" << endl);
      AFF_ERR("GET DATA", "Initialization file (" << fullfn << ") not found (" << hex << ecode << ")!");
      return ecode;
    }

    parse_pi(data,TRUE);

  //-[system data]------------------------------------------------------------
    AFF_LOG("    * getting system data .. ");

    if (sysdat.Length())
    {
      AFF_LOG("OKAY!" << endl);
      parse_pi(sysdat,TRUE);
    }
    else
    {
      AFF_LOG("FAIL: <none>" << endl);
    }

  //-[QUERY_STRING environment variable]--------------------------------------

    AFF_LOG("    * reading QUERY_STRING ... ");

    if (AFF_CALLTYPE == AFF_CALLTYPE_REMOTE && QUERY_STRING != NULL && *QUERY_STRING != '\0')
    {
      AFF_LOG("OK!" << endl << "    * QUERY_STRING = \"" << QUERY_STRING << 
              "\"" << endl);
      data = '&';
      data.add(QUERY_STRING);
      data.add(';');
      AFF_LOG("    * mod QUERY_STRING = \"" << data << "\"" << endl);
      parse_pi(data,FALSE);
    }
    else
    {
      AFF_LOG("FAIL: " <<
              ((AFF_CALLTYPE==AFF_CALLTYPE_LOCAL)?
              "LOCAL CALL-TYPE":"NOT FOUND") << endl);
    }

  //-[standard input stream]--------------------------------------------------

    AFF_LOG("    * standard input ... ");

    if (AFF_CALLTYPE == AFF_CALLTYPE_REMOTE && (clen = CONTENT_LENGTH) > 0)
    {
      AFF_LOG("OK!" << endl);
      if (C7STR::buffer_size < 10) C7STR::buffer_size = 10;
      data.alloc(clen);
      cin.get(data(), clen);
      parse_pi(data,FALSE);
    }
    else
    {
      AFF_LOG("FAIL: " <<
              ((AFF_CALLTYPE==AFF_CALLTYPE_LOCAL)?
              "LOCAL CALL-TYPE":"NOT PRESENT") << endl);
    }

  //-[specified input files]--------------------------------------------------

    AFF_LOG("    * specific file(s) ... ");

    anyinp = FALSE;

    //-[beg sub-loop of process items]----------------------------------------
    for (i = 0; i < sz_pis && ecode == C7EC_OK; i++)
    {

      if (pis[i].id == AFF_PI_INC)
      {
        //_[beg sub-loop of process item values]______________________________
        for (si = 0;                            // reset sub-index value
             si < AFF_MAX_PI_VALUES      &&     // while its less than da max
             pis[i].value[si].Length()>0 &&     // while there are values there
             ecode == C7EC_OK             ;     // while no errors have occured
             si++)                              // increment sub-index value
        {
          if (sysdir.Length() && pis[i].value[0].indexOf('/') < 0 && pis[i].value[0].indexOf('\\') < 0)
            fullfn = sysdir;
          else if (pis[i].value[0].indexOf('/') < 0 && pis[i].value[0].indexOf('\\') < 0)
            fullfn = "\\";
          else
            fullfn = "";
          fullfn.add(pis[i].value[0]);

          AFF_LOG(((anyinp==TRUE)?"    * reading inc file ... ":"") << 
                  '[' << fullfn << ']');

          if ((ecode = data.load(fullfn)) == C7EC_OK)
          {
            AFF_LOG(" OK" << endl);
            parse_pi(data,FALSE);
          }
          else
          {
            AFF_LOG(" FAIL (" << hex << ecode << dec << ")!" << endl);
            AFF_ERR("GET DATA","Could not open file to include: " <<
                    fullfn << '(' << ecode << ')');
          }
          anyinp = TRUE;
        }
        //_[end sub-loop of process item values]______________________________
      }
    }
    //-[end sub-loop of process items]----------------------------------------

    AFF_LOG(((anyinp==FALSE)?"FAIL: NONE":" ") << endl);

  //-[done getting data]------------------------------------------------------

  return 0;
}

//-[parse_pi]-----------------------------------------------------------------
//
FATHOMec aff_program::parse_pi(C7STR &data, bool master)
{
  FATHOMi sp;               // starting point for searches
  FATHOMi nxt;              // next point for variable
  FATHOMi bi, ei;           // beginning/ending indices
  uint32 id;                // stored identification
  FATHOMec ecode=C7EC_OK;   // error code
  C7STR encode;             // for encoding identification
  C7STR src;                // source (after url decoding)

  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  src.alloc(data.Length());
  fbc_url_decode(data(),src());

  nxt = src.indexOf('&',0);

  while (nxt >= 0 && ecode == C7EC_OK)
  {
    sp = nxt;

    if (sz_pis >= AFF_MAX_PROCESS_ITEMS)
    {
      AFF_ERR("PARSE PI","Maximum process items reached (" <<
              AFF_MAX_PROCESS_ITEMS << ")!");
      break;
    }

    sp++;
    ei = src.indexOf('=',sp);
    nxt = src.indexOf('&',sp);

    if ((nxt < ei && nxt > sp) || (ei < 0)) // no equals sign following '&'
    {
      if (sp > src.Length() - 4) src.add("    ");
      AFF_ERR("PARSE PI","Stray ampersand at index " << sp <<
              " (\""   << (char)c7pcha(src[sp-4]) << (char)c7pcha(src[sp-3])
                       << (char)c7pcha(src[sp-2]) << (char)c7pcha(src[sp-1])
                       << (char)c7pcha(src[sp])
                       << (char)c7pcha(src[sp+1]) << (char)c7pcha(src[sp+2])
                       << (char)c7pcha(src[sp+3]) << (char)c7pcha(src[sp+4])
              << "\")!");
      break;
    }
    else if (ei == sp)      // equals sign immediately follows '&', i.e. "&="
    {
      if (sp > src.Length() - 4) src.add("    ");
      AFF_ERR("PARSE PI","No variable name at index " << sp <<
              " (\""   << (char)c7pcha(src[sp-4]) << (char)c7pcha(src[sp-3])
                       << (char)c7pcha(src[sp-2]) << (char)c7pcha(src[sp-1])
                       << (char)c7pcha(src[sp])
                       << (char)c7pcha(src[sp+1]) << (char)c7pcha(src[sp+2])
                       << (char)c7pcha(src[sp+3]) << (char)c7pcha(src[sp+4])
              << "\")!");
      break;
    }

    encode.getpart(src,sp,ei);
    ei++;
    while (encode.Length() < 4)
      encode.add(' ');
    id = AFF_PID(encode[0],encode[1],encode[2],encode[3]);

    AFF_LOG("      & process item: "
            << encode[0] << encode[1] << encode[2] << encode[3] << " (" << hex
            << id << dec << "): ");

    if (pis[sz_pis].set(id,master) != C7EC_OK)
    {
      AFF_LOG("BAD ID!" << endl);
      AFF_WRN("PARSE PI","Invalid process item identification (" <<
              encode << ")!");
    }
    else
    {
      bi = 0;
      ecode = C7EC_OK;

      while (src[ei] != ';'    && // end at a semi-colon
             src[ei] != '&'    && // end at an ampersand
             ei < src.Length() && // end if exceeded length
             ecode == C7EC_OK)    // end if error happens
      { //-[beg loop looking for this process item's values]------------------

        if (src[ei] == ',')
        {
          AFF_LOG('[' << pis[sz_pis].value[bi] << ']');
          bi++;
          if (bi == AFF_MAX_PI_VALUES)
          {
            AFF_WRN("PARSE PI","Too many values for '" << encode << "'!");
            ecode = C7EC_ERROR;
          }
        }
        else
          pis[sz_pis].value[bi].add(src[ei]);
        ei++;

      } //-[end loop looking for this process item's values]------------------

      ecode = C7EC_OK;

      if (pis[sz_pis].value[bi].Length())
      {
        AFF_LOG('[' << pis[sz_pis].value[bi] << ']');
      }
      else if (bi == 0)
      {
        AFF_ERR("PARSE PI","No values for '" << encode << "'!");
        ecode = C7EC_ERROR;
      }

      AFF_LOG(endl);
      sz_pis++;
    }
  } 

  return C7EC_OK;
}

//-[enable_logfile]-----------------------------------------------------------
//
FATHOMec aff_program::enable_logfile(char *logf)
{
  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  if (logging == TRUE) return C7EC_OK;
  logging = TRUE;

  if (logf == NULL)
    lf.free();
  else
    lf = logf;

  if (sysdir.Length())
    tmpll = sysdir;
  else
    tmpll = "\\";
  tmpll.add(lf);

  logfile.open(tmpll(), ios::app);
  
  if (logfile.fail())
  {
    AFF_ERR("LOG FILE", "Unable to open log file (" << lf << ") for append!");
    lf.free();
    return C7EC_CANT_OPEN_FILE;
  }
  else
  {
    AFF_GET_SYS_TIME;
    logfile << endl << "AFF Log File (" << systime << ')' << endl
            << "----------------------------------------"
            << "----------------------------------------" << endl;
  }

  logfile.close();

  return C7EC_OK;  
}

//-[enable_errlist]-----------------------------------------------------------
//
FATHOMec aff_program::enable_errlist(char *errl)
{
  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  if (listing == TRUE) return C7EC_OK;
  listing = TRUE;

  if (errl == NULL)
    el.free();
  else
    el = errl;

  if (sysdir.Length())
    tmpll = sysdir;
  else
    tmpll = "\\";
  tmpll.add(el);

  errlist.open(tmpll(), ios::app);
  
  if (errlist.fail())
  {
    el.free();
    return C7EC_CANT_OPEN_FILE;
  }

  errlist.close();
  return C7EC_OK;
}

//-[reg_etype]----------------------------------------------------------------
//
FATHOMec aff_program::reg_etype(int32 id, char *nam, byte app, char *ext, char *dir, bool calc)
{
  FATHOMec retval;

  if (status == AFF_ERROR) return C7EC_UNABLE_TO_CONTINUE;

  AFF_LOG
  (
    "  ! Registering EType " << nam << "(0x" << hex << id << dec << ",0x" <<
    hex << (FATHOMi)app << dec << ",*." << ext << ',' << dir << ") ... "
  );

  if (sz_et == AFF_MAX_ETYPES)
  {
    AFF_LOG("FAIL (table full)!" << endl);
    AFF_WRN("REG ETYPE","Element types table full (" << nam << ")!");
    return C7EC_FAIL;
  }

  if ((retval = et[sz_et].set(id,nam,app,ext,dir,calc)) != C7EC_OK)
  {
    AFF_LOG("FAIL (" << hex << (FATHOMi)retval << dec << ")!" << endl);
    AFF_WRN("REG ETYPE","Calling element type class function 'set' failed ("
            << hex << (FATHOMi)retval << dec << ")!");
    return retval;
  }

  AFF_LOG("OKAY (" << sz_et << ")!" << endl);
  sz_et++;
  return C7EC_OK;
}

FATHOMi aff_program::handle_etype(int32 id)
{
  FATHOMi srch = 0;

  while (srch < sz_et)
  {
    if (et[srch].id == id)
      break;
    srch++;
  }

  return srch;
}

FATHOMec aff_program::check_security()
{
  AFF_LOG("  -[security]"
          "----------------------------------"
          "---------------------------------"
          << endl);

  if (AFF_CALLTYPE == AFF_CALLTYPE_LOCAL)
  {
    AFF_LOG("    $ Local Call-Type" << endl);
    return C7EC_OK;
  }

  C7STR server_ip;
  server_ip = CALLER_IP;

  AFF_LOG("  $ Gathering Caller IP: " << server_ip << endl);

  if (!rip.Length())
  {
    AFF_LOG("  $ Full-Access Granted: No Required IPs Specified" << endl);
    return C7EC_OK;
  }

  if (rip.indexOf(server_ip) >= 0 || server_ip.indexOf(rip) >= 0)
  {
    AFF_LOG("  $ Server is Authorized" << endl);
    return C7EC_OK;
  }

  AFF_LOG("  $ Unauthorized server!" << endl);
  AFF_ERR("SECURITY",
          "Unauthorized server (" << server_ip << ") attempted access!");

  return C7EC_UNABLE_TO_CONTINUE;
}

FATHOMec aff_program::chexec()
{
  if (!strcmp(aff_mode, "{lite  }"))
    exec_opts = AFFEM_NORMAL_PROCESS;
  else if (!strcmp(aff_mode, "{normal}"))
    exec_opts = AFFEM_INITIAL_ERROR_LISTING |
                AFFEM_NORMAL_PROCESS;
  else if (!strcmp(aff_mode, "{debug }"))
    exec_opts = AFFEM_INITIAL_ERROR_LISTING |
                AFFEM_INITIAL_PROCESS_LOGGING |
                AFFEM_NORMAL_PROCESS;
  else if (!strcmp(aff_mode, "{hdebug}"))
    exec_opts = AFFEM_INITIAL_ERROR_LISTING |
                AFFEM_INITIAL_PROCESS_LOGGING |
                AFFEM_INITIAL_OUTPUT |
                AFFEM_NORMAL_PROCESS;
  else if (!strcmp(aff_mode, "{fileio}"))
    exec_opts = AFFEM_RUN_IO_TEST;
  else if (!strcmp(aff_mode, "{cgitxt}"))
    exec_opts = AFFEM_RUN_CGITXT_TEST;
  else if (!strcmp(aff_mode, "{cgijpg}"))
    exec_opts = AFFEM_RUN_CGIJPG_TEST;
  else
    AFF_ERR("CHEXEC", "Bad execution mode [" << aff_mode << "]");


  return C7EC_OK;
}

FATHOMec aff_program::info_head()
{
  static int ran_head = 0;
  
  if (!ran_head)
  {
    AFF_GET_SYS_TIME;

    cout << "Content-Type: text/html\n\n"
                "<HTML>"
                  "<HEAD><TITLE>AFF Output: " << systime <<  "</TITLE></HEAD>"
                  "<CENTER><H1>AFF</H1></CENTER>"
                  "<PRE>";

    ran_head = 1;
  }

  return C7EC_OK;
}

FATHOMec aff_program::info_foot()
{
  static int ran_foot = 0;

  if (!ran_foot)
  {
    AFF_GET_SYS_TIME;

    cout << "</PRE><BR><BR><I>" << systime << "</I></HTML>";

    ran_foot = 1;
  }

  return C7EC_OK;
}

FATHOMec aff_program::command_line(char *cm)
{
  C7STR cmd;
  int cmdi = 0;

  cmd.is(cm);
  cmd.case_flag(0);

  if (cmd == "mode") cmdi = 1;
  else if (cmd == "sysd") cmdi = 2;
  else if (cmd == "info") cmdi = 10000;
  
  if (cmdi)
  {
    cout << PROG_NAME << " v" << PROG_VERS << " (" << CINF_ENV << ":" 
         << CINF_NAM << ")" << endl
         << PROG_AUTH << ", " << PROG_DATE << endl << endl;

    switch (cmdi)
    {
      case 1:
        cout << "EXECUTION MODE: " << aff_mode << endl;
        break;
      case 2:
        cout << "SYSTEM DIRECTORY: " << get_sysd() << endl;
        break;
      default:
        break;
    }
  }
  else
    return C7EC_FAIL;

  return C7EC_OK;  
}